#include <minix/config.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sgtty.h>
#include <sys/fdio.h>
#include <fcntl.h>
#include <stdio.h>

_PROTOTYPE(int main, (int argc, char **argv));

main(argc, argv)
int argc; char **argv;
{
	struct fparam fdio;
	int sfd;

	int  set_param	 = 0;
	int  sector_size = -1;	/* sector size in bytes */
	char sector_0    = -1;	/* first sector on a track */
	char nr_sectors  = -1;	/* sectors per track */
	char cylinder_0  = -1;	/* cylinder offset */
	char nr_cylinders= -1;	/* number of cylinders */
	char nr_sides    = -1;	/* :1 single/double sided */
	char density     = -1;	/* :3 density code */
	char stepping    = -1;	/* :1 double stepping */
	extern char *optarg;
	extern int optind;
	char dev1[64], dev2[64];
	register char c;
	static char usage[] = "Usage: %s [-l sector size]\n\
        [-s [sector 0,]sectors]\n\
        [-c [cylinder 0,]cylinders]\n\
        [-1|-2] [-D (DD|QD|HD|DD5)] [-S n|d] raw_device\n";

	while ((c = getopt(argc, argv, "l:s:c:12D:S:?")) != EOF) {
		switch(c) {
			case 'l':
				sector_size = atoi(optarg);
				break;

			case 's':
				if (strchr(optarg, ',')) {
					sector_0 = atoi(optarg);
					optarg = strchr(optarg, ',') + 1;
				}
				nr_sectors = atoi(optarg);
				break;

			case 'c':
				if (strchr(optarg, ',')) {
					cylinder_0 = atoi(optarg);
					optarg = strchr(optarg, ',') + 1;
				}
				nr_cylinders = atoi(optarg);
				break;

			case '1':
				nr_sides = 0;
				break;

			case '2':
				nr_sides = 1;
				break;

			case 'D':
				if (!strcmp(optarg, "DD"))
					density = DD;
				else if (!strcmp(optarg, "QD"))
					density = QD;
				else if (!strcmp(optarg, "HD"))
					density = HD;
				else if (!strcmp(optarg, "DD5"))
					density = DD5;
				else {
					fprintf(stderr, usage, argv[0]);
					exit(1);
				}
				break;

			case 'S':
				if (!strcmp(optarg, "d"))
					stepping = DSTP;
				else if (!strcmp(optarg, "n"))
					stepping = NSTP;
				else {
					fprintf(stderr, usage, argv[0]);
					exit(1);
				}
				break;

			default:
				fprintf(stderr, usage, argv[0]);
				exit(1);
		}
	}

	if (!argv[optind]) {
		fprintf(stderr, usage, argv[0]);
		exit(1);
	}

	*dev2 = '\0';
	if (!strncmp(argv[optind], "/dev/r", 6))
		strcpy(dev1, argv[optind]);
	else if (!strncmp(argv[optind], "/dev/dsk/", 9))
		sprintf(dev1, "/dev/rdsk/r%s", argv[optind] + 9);
	else if (!strncmp(argv[optind], "/dev/", 5))
		sprintf(dev1, "/dev/r%s", argv[optind] + 5);
	else if (*argv[optind] == 'r') {
		sprintf(dev1, "/dev/%s", argv[optind]);
		sprintf(dev2, "/dev/rdsk/%s", argv[optind]);
	} else if (*argv[optind] != '/') {
		sprintf(dev1, "/dev/r%s", argv[optind]);
		sprintf(dev2, "/dev/rdsk/r%s", argv[optind]);
	} else
		strcpy(dev1, argv[optind]);

	if (!*dev2)
		strcpy(dev2, dev1);

	if (((sfd = open(dev1, O_RDONLY)) < 0  &&
	     (sfd = open(dev2, O_RDONLY)) < 0) ||
	     ioctl(sfd, DIOGETP, (struct sgttyb *)&fdio) < 0) {
		perror(argv[0]);
		fprintf(stderr, usage, argv[0]);
		exit(1);
	} else {
		if (nr_cylinders != -1) {
			set_param = 1;
			fdio.nr_cylinders = nr_cylinders;
		}
		if (sector_size != -1) {
			set_param = 1;
			fdio.sector_size = sector_size;
		}
		if (sector_0 != -1) {
			set_param = 1;
			fdio.sector_0 = sector_0;
		}
		if (nr_sectors != -1) {
			set_param = 1;
			fdio.nr_sectors = nr_sectors;
		}
		if (cylinder_0 != -1) {
			set_param = 1;
			fdio.cylinder_0 = cylinder_0;
		}
		if (nr_cylinders != -1) {
			set_param = 1;
			fdio.nr_cylinders = nr_cylinders;
		}
		if (nr_sides != -1) {
			set_param = 1;
			fdio.nr_sides = nr_sides;
		}
		if (density != -1) {
			set_param = 1;
			fdio.density = density;
		}
		if (stepping != -1) {
			set_param = 1;
			fdio.stepping = stepping;
		}
	}

	if (set_param && (ioctl(sfd, DIOSETP, (struct sgttyb *)&fdio) < 0)) {
		perror(argv[0]);
		fprintf(stderr, usage, argv[0]);
		exit(1);
	}

	printf("sector_size  = %d\n", fdio.sector_size);
	printf("sector_0     = %d\n", fdio.sector_0);
	printf("nr_sectors   = %d\n", fdio.nr_sectors);
	printf("cylinder_0   = %d\n", fdio.cylinder_0);
	printf("nr_cylinders = %d\n", fdio.nr_cylinders);
	printf("nr_sides     = %d\n", fdio.nr_sides + 1);
	printf("density      = %d\n", fdio.density);
	printf("stepping     = %d\n", fdio.stepping);
	exit(0);
}
