/*
 * This file contains support for the Motorola 68881/2 FPA chip.
 * It is assumed that the kernel itself (syscall) never uses the FPP.
 *
 * The FPP state frame is always saved/restored. If the FPP state frame
 * is a NULL frame the FPP register set is not saved/restored.
 *
 * The entry points into this file are
 *	fpp_new_state:	initialize the FPP for a new process
 *	fpp_save:	save the FPP state frame and the FPP register set
 *	fpp_restore:	restore the FPP state frame and the FPP register set
 */

#include "kernel.h"
#include "proc.h"

#ifdef FPP
/*===========================================================================*
 *				fpp_new_state				     *
 *===========================================================================*/
PUBLIC void fpp_new_state(rp)
struct proc *rp;
{
  rp->p_fsave.cpu_state.i_format     =
  rp->p_fsave.state_frame.frame_type =
  rp->p_fsave.state_frame.frame_size =
  rp->p_fsave.state_frame.reserved   = 0;
}


/*===========================================================================*
 *				fpp_save				     *
 *===========================================================================*/
PUBLIC void fpp_save(rp, p)
struct proc *rp; struct cpu_state *p;
{
  /* Do not save state for tasks */
  if (rp->p_nr < 0)
	return;

  _fppsave(&(rp->p_fsave.state_frame));
  if (rp->p_fsave.state_frame.frame_type)	{
	/* only dump registers if not null */
	_fppsavereg(&(rp->p_fsave.fpp_model));
	rp->p_fsave.cpu_state = *p;
  }
}


/*===========================================================================*
 *				fpp_restore				     *
 *===========================================================================*/
PUBLIC struct cpu_state *fpp_restore(rp)
struct proc *rp;
{
  register struct cpu_state *cp = (struct cpu_state *)NULL;

  if (rp->p_nr < 0)
	return((struct cpu_state *) NULL);

  if (rp->p_fsave.state_frame.frame_type) {
	/* only restore when !null state */
	_fpprestreg(&(rp->p_fsave.fpp_model));
	if ((rp->p_fsave.cpu_state.i_format & 0xf000) == 0x9000)
		cp = &(rp->p_fsave.cpu_state);
  }
  _fpprestore(&(rp->p_fsave.state_frame));
  return(cp);
}
#endif
