#include <minix/config.h>
#if (CHIP == M68000)

#include "const.h"

|
| Only the registers that are considered scratch by the compiler used
| to compile the kernel are saved. Other registers are saved by the
| prolog generated by the kernel compiler. NOTE: d0, a0 and a1 are
| used by the asm interrupt handler. At least these *must* be saved.
|

#define FREEREGS d0-d1/a0-a1
#define SPACEREGS 16

| public labels
|
.globl	_lock
.globl	_unlock
.globl	_restore
.globl	_reboot
.globl	_test_and_set
.globl	_get_mem_size
.globl	_sizes
.globl	_invicache
.globl	_invdcache
.globl	_proctyp
.globl	_fpptyp
.globl	_mmutyp

|
| external references
|
.globl	_clock_handler
.globl	_held_head
.globl	_k_reenter
.globl	_proc_ptr
.globl	_rdy_head
.globl	_cstart
.globl	_trap
.globl	_none
.globl	_rupt
.globl	_timint
.globl	_dmaint
.globl	_siaint
.globl	_aciaint
.globl	_piaint
.globl	_iob
.globl	_sys_call
.globl	_lock_pick_proc


|
| offsets into a proc table entry (see proc.h)
|
#define sava6	56
#define savsp	60
#define savpc	64
#define savsr	68
#define savtt	77

.text

.long	0		|   0	0 reset initial sp; cannot be written on ATARI

.long	start		|   1	4 reset initial pc
.long	buserr		|   2	8 bus error
.long	adrerr		|   3	c adress error
.long	illtrp		|   4  10 illegal instruction
.long	zertrp		|   5  14 zero divide
.long	chktrp		|   6  18 chk/chk2 instruction
.long	trptrp		|   7  1c trapcc/trapv instruction
.long	prvtrp		|   8  20 privilege violation
.long	trc		|   9  24 trace
.long	atrp		|  10  28 line 1010 emulator
.long	ftrp		|  11  2c line 1111 emulator
.long	trp		|  12  30 reserved
.long	cpvtrp		|  13  34 coprocessor protocol violation
.long	fortrp		|  14  38 format error
.long	trp		|  15  3c uninitialized interrupt

.long	trp, trp, trp, trp, trp, trp, trp, trp		| 16 - 23 reserved

.long	intr0		|  24  60 spurious interrupt
.long	intr1		|  25  64 level 1 interrupt vector
.long	intr2		|  26  68 level 2 interrupt autovector
.long	intr3		|  27  6c level 3 interrupt autovector
.long	intr4		|  28  70 level 4 interrupt autovector
.long	intr5		|  29  74 level 5 interrupt autovector
.long	intr6		|  30  78 level 6 interrupt autovector
.long	intr7		|  31  7c level 7 interrupt autovector
.long	sys		|  32  80 trap #0, used for system calls
			|  33 - 47 trap #1-15
.long	trpntrp, trpntrp, trpntrp, trpntrp, trpntrp, trpntrp, trpntrp
.long	trpntrp, trpntrp, trpntrp, trpntrp, trpntrp, trpntrp, trpntrp, trpntrp
			|  48 - 55  fpcp errors
.long	fpcptrp, fpcptrp, fpcptrp, fpcptrp, fpcptrp, fpcptrp, fpcptrp, fpcptrp
			|  56 - 58  pmmu errors
.long	pmmutrp, pmmutrp, pmmutrp
			|  59 - 63  reserved
.long	trp, trp, trp, trp, trp
|  rest are user-defined vectors
| Asynchronous interrupts
.long	pia		| 64 par printer busy
.long	sia4		| 65 rs232 DCD
.long	sia4		| 66 rs232 CTS
.long	iob3		| 67 not handled (blitter done)
.long	tim3		| 68 timer D
.long	clk		| 69 timer C
.long	acia		| 70 keyboard/MIDI
.long	dma		| 71 fdc/acsi
.long	tim1		| 72 timer B
.long	sia3		| 73 rs232 transmit error
.long	sia2		| 74 rs232 transmit buffer empty
.long	sia1		| 75 rs232 receive error
.long	sia0		| 76 rs232 receive buffer full
.long	tim0		| 77 timer A
.long	iob6		| 78 not handled (rs232 RI signal)
.long	iob7		| 79 not handled (monochrome monitor detect)
.long	iob8		| 80 not handled (unused)
.long	iob9		| 81 not handled (unused)
.long	non		| 82 scc dma interrupt
.long	non		| 83 ring indicator scc b
.long	tim7		| 84 timer D
.long	tim6		| 85 timer C
.long	non		| 86 reserved (FDC Drive Ready ? HO )
.long	scsidma		| 87 scsi dma interrupt
.long	tim5		| 88 timer B
.long	sia8		| 89 rs232 transmit error
.long	sia7		| 90 rs232 transmit buffer empty
.long	sia6		| 91 rs232 receive error
.long	sia5		| 92 rs232 receive buffer full
.long	tim4		| 93 timer A
.long	non		| 94 rtc irq (cleared by reading rtc)
.long	scsi		| 95 scsi controller irq
| Asynchronous interrupts end here. rest is unused
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non
.long	non, non, non, non, non, non, non, non

| startup code, just after interrupt vectors
| call _cstart and do process switch when it returns
|
| ATTENTION: DO NOT TAMPER WITH D1,A2!
| The loader passes a pointer to the config data in A2.
|
start:
|	reset
	movew	#0x2700,sr
	movel	#k_stktop,sp	| load sp so exceptions can be stored

	clrb	_fpptyp		| assume no fpu

| first determine the mmu type
	movel	0x2c,a0		| save bus error exception handler
	movel	#proctype,0x2c	| install own error handler
	clrb	_mmutyp		| no mmu
	pflusha
	moveb	#1,_mmutyp	| assume it is a 68851. If it is a 68030
				| instead we will detect this later on
proctype:
	movel	a0,0x2c		| restore old invalid instruction exception handler

| now determine proc type
| warning: the detection code for 68040 has never been tested!!!
	movel	0x10,a0		| save invalid instruction exception handler
	movel	#fputype,0x10		| install own error handler
| It is at least a 68000/68008 
	clrb	_proctyp
| see if it is an 68010 or up
	move	ccr,d0
| since we came here the move instruction is legal so 68010 or up
	moveb	#1,_proctyp
| see if it is 68020 or up. If so also flush and enable the instruction cache
	movel	#0x3919,d0
	movecl	d0,cacr
	moveb	#2,_proctyp
| see if it is 68030 or 68040
| this is done by reading the cacr register back 
	movecl	cacr,d0
| if d0 = 0 we have movec and cacr, but we have not enabled a cache
| with movec.l d0,cacr. So it must be an 68040
	tstw	d0
	beq	got_68040
	andw	#0xff00,d0
| if no there are no bits set in the second byte then we have no data
| cache and thus an 68020
	beq	fputype
	moveb	#3,_proctyp
| if it is an 68030 then also set mmutyp
	moveb	#3,_mmutyp
	bra	fputype
got_68040:
	movel	#0x80008000,d0		| enable both caches
	movecl	d0,cacr
	moveb	#4,_proctyp
	moveb	#4,_mmutyp
	moveb	#4,_fpptyp
| even if the 68040 has a fpu on board check if a coprocessor
| with the full fpp instruction set is present
fputype:
	movel	a0,0x10		| restore old invalid instruction exception handler

| now determine fpp type
	movel	0x2c,a0		| save bus error exception handler
	movel	#got_types,0x2c		| install own error handler
	movel	#0,a7@-
	frestore a7@+
	fsave	a7@-
| if we got here there is a fpp chip
| discriminate between 68881 and 68882
	cmpb	#0x18,a7@(1)		| if this byte contains 0x18 it is a 68881
| if it contains 0x38 it is a 68882
	bne	got_68882
	moveb	#1,_fpptyp
	bra	got_types
got_68882:
	moveb	#2,_fpptyp
got_types:
	movel	a0,0x2c		| restore old invalid instruction exception handler

	movel	#k_stktop,sp		| forget the exception data on the stack

	movel	8,a0		| save old bus error exception handler
	movel	#berr,8		| install own error handler

| the next two instructions enable the cache and set the processor speed to
| 16 Mhz. This only works on Mega STe computers and causes a bus error on
| "normal" ST-s. In that case the instructions are just skipped
	bset	#0,0xffff8e21		| enable Mega STe cache
	bset	#1,0xffff8e21		| Mega STe 16 Mhz speed
berr:
	movel	#k_stktop,sp		| forget the exception data on the stack
	movel	a0,8		| restore old bus error exception handler

| the following code could be used to clear the bss area
|	move.l	#_end,d0
|	move.l	#_edata,a0
|	bra	L2
|L1:
|	move.w	#0,(a0)+
|L2:
|	cmp.l	d0,a0
|	bcs	L1

| kernel configuration from loader/menu
	movew	d1,sp@-		| D1 = boot parameter length
	clrw	sp@-
	movel	a2,sp@-		| A2 = boot parameter pointer
	jsr	_cstart
	addl	#8,sp
	moveb	#0x0,0xffff8240	| Color register ? HO
	bra	restart

|
| the vector handlers for synchronous traps. save full context.
|
buserr:
	moveb	#2,vectnum
	tstb	_proctyp
	bne	gentrp
	addl	#8,sp		| remove extra context on 68000 processor
	bra	gentrp
adrerr:
	moveb	#3,vectnum
	tstb	_proctyp
	bne	gentrp
	addl	#8,sp		| remove extra context on 68000 processor
	bra	gentrp
trp:
	moveb	#12,vectnum
	bra	gentrp
illtrp:
	moveb	#4,vectnum
	bra	gentrp
zertrp:
	moveb	#5,vectnum
	bra	gentrp
chktrp:
	moveb	#6,vectnum
	bra	gentrp
trptrp:
	moveb	#7,vectnum
	bra	gentrp
prvtrp:
	moveb	#8,vectnum
	bra	gentrp
atrp:
	moveb	#10,vectnum
	bra	gentrp
ftrp:
	moveb	#11,vectnum
	bra	gentrp
cpvtrp:
	moveb	#13,vectnum
	bra	gentrp
fortrp:
	moveb	#14,vectnum
	bra	gentrp
trpntrp:
	moveb	#33,vectnum
	bra	gentrp
fpcptrp:
	moveb	#48,vectnum
	bra	gentrp
pmmutrp:
	moveb	#56,vectnum
	bra	gentrp

gentrp:
	bsr	save
	jsr	_trap
	bra	restart
trc:
	moveb	#9,vectnum
	btst	#5,sp@		| tracing through trap?
	beq	gentrp		| No, do normal trace processing
	rte			| do not trace; execute system call
non:
	lsrw	sp@(6)		| HO
	lsrw	sp@(6)		| HO
	moveb	sp@(7),vectnum	| HO
	bsr	save
	jsr	_none
	bra	restart

sys:
	moveb	#32,vectnum
	bsr	save		| d0, d1 and a0 not modified
	bsr	_unlock		| do not lock during sys_call
	movel	a0,sp@-		| m_ptr
	movew	d1,sp@-		| src_dest
	movew	d0,sp@-		| SEND/RECEIVE/BOTH
| warning: _sys_call may change _proc_ptr
	jsr	_sys_call		| sys_call(func,src_dest,m_ptr)
	swap	d0
	clrw	d0
	swap	d0
	movel	d0,a6@		| a6 is set by save to _proc_ptr
	addl	#8,sp

	movew	#0x2700,sr
	tstb	pick_proc		| _proc_ptr changed in a interrupt routine?
	beq	1f
	sf	pick_proc		| clear pick_proc
	jsr	_lock_pick_proc		| _proc_ptr changed -> call _lock_pick_proc
1:	bra	restart

intr0:
intr1:
intr3:
intr5:
intr6:
intr7:
	bsr	save
	jsr	_rupt		| call service routine
	bra	restart
intr2:
| this one does not affect task switching
| must be avoided because it comes so frequent
| execute spl2()
	orw	#0x200,sr
	rte	
intr4:
| if _progvdu is set, call interrupt handler to load
| new values in the video registers, and clear _progvdu.
| if _progvdu is not set, just rte
	tstw	_progvdu
	bne	dovbl
	rte	

dovbl:	moveml	FREEREGS,sp@-
	movew	#0,_progvdu
	movew	#0,d0
	movel	#_vbl,a0
	bra	genint

|
| The vector handlers for asynchronous traps. context saved only if necessary.
|
iob3:
	moveml	FREEREGS,sp@-
	movel	#3,d0
	movel	#_iob,a0
	bra	genint
iob6:
	moveml	FREEREGS,sp@-
	movel	#6,d0
	movel	#_iob,a0
	bra	genint
iob7:
	moveml	FREEREGS,sp@-
	movel	#7,d0
	movel	#_iob,a0
	bra	genint
iob8:
	moveml	FREEREGS,sp@-
	movel	#8,d0
	movel	#_iob,a0
	bra	genint
iob9:
	moveml	FREEREGS,sp@-
	movel	#9,d0
	movel	#_iob,a0
	bra	genint

tim0:
	moveml	FREEREGS,sp@-
	movel	#0,d0
	movel	#_timint,a0
	bra	genint
tim1:
	moveml	FREEREGS,sp@-
	movel	#1,d0
	movel	#_timint,a0
	bra	genint
tim3:
	moveml	FREEREGS,sp@-
	movel	#3,d0
	movel	#_timint,a0
	bra	genint
tim4:
	moveml	FREEREGS,sp@-
	movel	#4,d0
	movel	#_timint,a0
	bra	genint
tim5:
	moveml	FREEREGS,sp@-
	movel	#5,d0
	movel	#_timint,a0
	bra	genint
tim6:
	moveml	FREEREGS,sp@-
	movel	#6,d0
	movel	#_timint,a0
	bra	genint
tim7:
	moveml	FREEREGS,sp@-
	movel	#7,d0
	movel	#_timint,a0
	bra	genint

clk:
	subw	#1,clkcnt
	beq	cont
	rte	
cont:
	movew	#4,clkcnt
	moveml	FREEREGS,sp@-
	movel	#_clock_handler,a0
	bra	genint
dma:
	moveml	FREEREGS,sp@-
	movel	#_dmaint,a0
	bra	genint
scsidma:
	moveml	FREEREGS,sp@-
	movel	#_scsidmaint,a0
	bra	genint
scsi:
	moveml	FREEREGS,sp@-
	movel	#_scsiint,a0
	bra	genint
acia:
	moveml	FREEREGS,sp@-
	movel	#_aciaint,a0
	bra	genint
sia0:
	moveml	FREEREGS,sp@-
	movel	#0,d0
	movel	#_siaint,a0
	bra	genint
sia1:
	moveml	FREEREGS,sp@-
	movel	#1,d0
	movel	#_siaint,a0
	bra	genint
sia2:
	moveml	FREEREGS,sp@-
	movel	#2,d0
	movel	#_siaint,a0
	bra	genint
sia3:
	moveml	FREEREGS,sp@-
	movel	#3,d0
	movel	#_siaint,a0
	bra	genint
sia4:
	moveml	FREEREGS,sp@-
	movel	#4,d0
	movel	#_siaint,a0
	bra	genint
sia5:
	moveml	FREEREGS,sp@-
	movel	#5,d0
	movel	#_siaint,a0
	bra	genint
sia6:
	moveml	FREEREGS,sp@-
	movel	#6,d0
	movel	#_siaint,a0
	bra	genint
sia7:
	moveml	FREEREGS,sp@-
	movel	#7,d0
	movel	#_siaint,a0
	bra	genint
sia8:
	moveml	FREEREGS,sp@-
	movel	#8,d0
	movel	#_siaint,a0
	bra	genint

pia:
	moveml	FREEREGS,sp@-
	movel	#_piaint,a0
	bra	genint

genint:
	addb	#1,_k_reenter		| from -1 if not reentering
	bne	3f
| NOTE: the following 4 instructions will not cause a 
|       context switch if a kernel process with pc < restart is
|	interrupted. Currently such a process is always 
|	busy with a trap or irq. If in the future this is
|	not the case for some reason (e.g. every task has its
|	own virtual address), then process switches may 
|	not be done (while required)
	btst	#5,sp@(SPACEREGS)		| check if we did not interrupt a
	beq	2f		| trap or irq, first test if the
| interrupted process is in user mode
| if so, no problem, jump to label 2
	cmpl	#restart,sp@(2+SPACEREGS)		| if we interupted a process in
	bge	2f		| supervisor mode check if the old pc
| was smaller than restart. 
| if not, no problem, jump to label 2
	addb	#1,_k_reenter		| else probably interrupted trap or irq
| so increment k_reenter twice because
| the interrupted routine maybe has not
| gotten to incrementing k_reenter
| we have to do it here since otherwise
| we might end up doing a save of an
| interrupted user process (due to a
| context switch) but we can
| not save both the usp and ssp
| incrementing k_reenter twice avoids
| the context switch!

| it is now clear that we are not
| a "level 0" interrupt service.
| only a "level 0" service may change
| proc_ptr.
	movel	_proc_ptr,sp@-		| save _proc_ptr
	movew	d0,sp@-		| push (potential) argument
	jsr	a0@		| call service routine
	addl	#2,sp		| pop argument
	movel	sp@+,d0

	movew	#0x2700,sr
	subb	#2,_k_reenter
	cmpl	_proc_ptr,d0
| if proc_ptr was changed, we have
| to restore it. on the other hand,
| the "level 0" service must be informed
| that a context-switch is now
| necessary/possible or we would
| eventually stay in idle for ever.
	beq	1f		| _proc_ptr changed?
	st	pick_proc		|    yes -> force pick_proc
	movel	d0,_proc_ptr		|           and restore _proc_ptr
1:
	moveml	sp@+,FREEREGS
	rte	

2:
	movel	_proc_ptr,a1		| save _proc_ptr
	movel	a1,proc_ptr
	movel	sp@(2+SPACEREGS),a1@(savpc)		| save pc for profiling in clock.c

	movew	d0,sp@-		| push argument
	jsr	a0@		| call service routine
	addl	#2,sp		| pop argument

	movew	#0x2700,sr
	subb	#1,_k_reenter
	movel	proc_ptr,d0		| _proc_ptr changed?
	cmpl	_proc_ptr,d0
	bne	1f
	tstb	pick_proc		| _proc_ptr changed in a interrupt routine?
	beq	2f
	sf	pick_proc		| clear pick_proc
	jsr	_lock_pick_proc		| _proc_ptr changed -> call _lock_pick_proc
	movel	proc_ptr,d0
	cmpl	_proc_ptr,d0
	beq	2f

1:	movel	_proc_ptr,proc_ptr
	movel	d0,_proc_ptr
	moveml	sp@+,FREEREGS

	bsr	save
	movel	proc_ptr,_proc_ptr
	bra	restart
2:
	moveml	sp@+,FREEREGS
	cmpl	#0,_held_head		| are there unserviced interrupts?
	bne	1f		| yes: let restart take care
	rte	

1:	bsr	save
	bra	restart
3:
	movel	_proc_ptr,sp@-		| save _proc_ptr
	movew	d0,sp@-		| push argument
	jsr	a0@		| call service routine
	addl	#2,sp		| pop argument
	movel	sp@+,d0

	movew	#0x2700,sr
	subb	#1,_k_reenter
	cmpl	_proc_ptr,d0
	beq	1f		| _proc_ptr changed?
	st	pick_proc		|    yes -> force pick_proc
	movel	d0,_proc_ptr		|           and restore _proc_ptr
1:
	moveml	sp@+,FREEREGS
	rte	

| perform task switch by save and restart
|
save:
	movew	#0x2700,sr
	movel	a6,sp@-
	movel	_proc_ptr,a6
	moveml	d0-d7/a0-a5,a6@
	movel	sp@+,a6@(sava6)		| a6
	movel	usp,a1
	btst	#5,sp@(4)		| test old S-bit
	beq	save_a1_ok
	lea	sp@(10),a1		| assume 68000
	tstb	_proctyp
	beq	save_a1_ok		| yes, it was really a 68000
| it was not a 68000, look at the frame number in the frame
	movel	#0,d6
	movew	sp@(10),d6		| format/offset word
	movel	#10,d7
	lsrl	d7,d6
	lea	framesizes,a1
	addl	d6,a1
	movel	a1@,a1
	addl	sp,a1
save_a1_ok:
	movel	a1,a6@(savsp)		| old sp: usp or ksp
	moveb	vectnum,a6@(savtt)		| trap type
	movel	sp@+,a1		| return address
	movew	sp@+,a6@(savsr)		| sr
	movel	sp@+,a6@(savpc)		| pc
	addb	#1,_k_reenter		| from -1 if not reentering

#ifdef FPP
| ifdef fpp, this could be inline to make it faster, also check to see
| if next process is a task, which does not use the fpp

	tstb	_fpptyp
	beq	no_fpp1
	moveml	d0/d1/a0/a1,sp@-		| save return addr
	lea	sp@(16),a0
	movel	a0,sp@-		| push address of format word
	movel	a6,sp@-		| push proc_ptr
	jsr	_fpp_save		| save fpp state
	addl	#8,sp
	moveml	sp@+,d0/d1/a0/a1
no_fpp1:
#endif
	lea	k_stktop,sp
	jmp	a1@

restart:
| Flush any held-up interrupts.
| This reenables interrupts, so the current interrupt handler may reenter.
| This does not matter, because the current handler is about to exit and no
| other handlers can reenter since flushing is only done when k_reenter == 0.

	movew	#0x2700,sr
	tstb	_k_reenter
	bne	over_call_unhold
	cmpl	#0,_held_head
	beq	over_call_unhold
	jsr	_unhold
over_call_unhold:
	jsr	_checksp
	movel	_proc_ptr,a6
	clrb	a6@(savtt)		| trap type
	clrb	vectnum
	movel	a6@(savsp),a0		| old sp: usp or ksp
	movel	#0,d7		| d7 == format word
	btst	#5,a6@(savsr)		| test old S-bit
	bne	L6		| jump if S-bit on
#if (SHADOWING == 0)
| pmmu load new map for user task
	movel	a0,sp@-		| save a0
	movel	a6,sp@-		| proc_ptr
	jsr	_pmmu_restore		| pmmu_restore(proc_ptr)
	movel	sp@+,a6
	movel	sp@+,a0		| restore a0
#endif
#ifdef FPP
	tstb	_fpptyp
	beq	no_fpp2
	movel	a0,sp@-		| save a0
	movel	a6,sp@-		| proc_ptr
	jsr	_fpp_restore		| fpp_restore(proc_ptr)
	movel	sp@+,a6
	movel	sp@+,a0		| restore a0
	tstl	d0		| restore stack frame?
	beq	1f
	movel	d0,a1
	movew	a1@+,d7		| set format word
	movel	a1@(8),sp@-
	movel	a1@(4),sp@-
	movel	a1@(0),sp@-		| copy stack frame
1:
no_fpp2:
#endif
	movel	a0,usp
	bra	L7
L6:
	movel	a0,sp
L7:
	tstb	_proctyp
	beq	L8
	movew	d7,sp@-		| set the format word
	jsr	_invicache
	jsr	_invdcache
L8:
	movel	a6@(savpc),sp@-		| pc
	movew	a6@(savsr),sp@-		| sr
	moveml	a6@,d0-d7/a0-a6
	movew	#0x2700,sr
	subb	#1,_k_reenter
	rte	

_invicache:
	cmpb	#1,_proctyp
	ble	invi1		| no instruction cache in 68000/010
	cmpb	#4,_proctyp
	bge	invi4		| different for 68040
	movecl	cacr,d0
	bset	#3,d0		| set the i cache clear bit
	movecl	d0,cacr
	rts	
invi4:
.word	0xf498		| cinv instruction cache
invi1:
	rts	

_invdcache:
	cmpb	#2,_proctyp
	ble	invd1		| no data cache in 68000/010/020
	cmpb	#4,_proctyp
	bge	invd4		| different for 68040
	movecl	cacr,d0
	bset	#11,d0		| set the d cache clear bit
	movecl	d0,cacr
	rts	
invd4:
.word	0xf458		| cinv data cache
invd1:
	rts	

_reboot:
	movel	4,a0
	jmp	a0@
_lock:
	movew	sr,d0
	movew	#0x2700,sr
	rts	
_restore:
	movew	sp@(4),sr
	rts	
_unlock:
	movew	#0x2200,sr		| spl2() to block HBL interrupt
	rts	
_test_and_set:
	movel	sp@(4),a0
	tas	a0@
	smi	d0
	extw	d0
	extl	d0
	rts	

_get_mem_size:
	cmpb	#2,_proctyp
	ble	nocache1
	cmpb	#4,_proctyp
	bge	cache4
	movel	#0x0019,d0		| disable data cache
	movecl	d0,cacr
	bra	nocache1
cache4:
	movel	#0x8000,d0		| disable data cache
	movecl	d0,cacr
nocache1:
	movel	sp@(4),a0		| start address
	link	a2,#0		| save stack pointer in case an exception occurs
	movel	8,a1		| save old bus error exception handler
	movel	#tstend,8		| install own error handler
next:
| offset 8 is used to avoid accessing unwritable memory if called with addr 0)
	movew	a0@(8),d0		| save value
	movew	#0x55aa,a0@(8)
	notw	a0@(8)
	cmpw	#0xaa55,a0@(8)
	bne	tstend
	movew	d0,a0@(8)		| restore original value
	addl	#0x40000,a0		| test every 256 k
	bra	next
tstend:
	movel	a1,8		| restore old bus error exception handler
	unlk	a2		| remove exception frame.
	cmpb	#2,_proctyp
	ble	end
	cmpb	#4,_proctyp
	bge	endcache4
	movel	#0x3919,d0		| re-enable data cache
	movecl	d0,cacr
	bra	end
endcache4:
	movel	#0x8080,d0		| re-enable data cache
	movecl	d0,cacr
end:
	subl	sp@(4),a0		| calculate offset
	movel	a0,d0
	rts	

.data

_sizes:
.word	0x526F, 0, 0, 0, 0, 0, 0, 0

conf:
.globl	_proc
.long	_proc
.globl	_keynorm
.long	_keynorm
.globl	_keyshft
.long	_keyshft
.globl	_keycaps
.long	_keycaps
.globl	_font8
.long	_font8
.globl	_font16
.long	_font16
.globl	_boot_parameters
.long	_boot_parameters
.long	0
.long	0

clkcnt:
.word	4

.skip	K_STACK_BYTES
k_stktop:

proc_ptr:		| old proc_ptr
.long	0
pick_proc:		| call pick_proc?
.byte	0
vectnum:		| trap/interrupt vector number
.byte	0
_proctyp:		| processor type, 0: 68000/68008, 1: 68010, 2: 68020,
|		  3: 68030, 4: 68040
.byte	0
_fpptyp:		| fpu processor type, 0: none, 1: 68881, 2: 68882, 4: 68040
.byte	0
_mmutyp:		| mmu type, 0: none, 1: 68851, 3: 68030, 4: 68040
.byte	0

| below are the frame sizes for 680x0 processors, x>0, size in bytes, +4 !!
| the sizes which are -1 are unused as far as I know.
framesizes:
.long	12, 12, 16, 16, 18, -1, -1, 64
.long	62, 24, 36, 96, 28, -1, -1, 38

.text
.globl	___main
___main:
	rts	

#endif
