/*-
 * Copyright (c) 2011-2014 Hans Ottevanger
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <math.h>

double modf(double x, double *ip)
{
	union {
		double d;
		struct {
			unsigned s:1;
			unsigned e:11;
			unsigned long f1:20;
			unsigned long f2:32;
		} b; /* works for big-endian only */
	} y;
	int nbits;

	y.d = x;

	nbits = 1023 + 52 - y.b.e;

	if (nbits > 52) {
		*ip = 0.0;
		return x;
	}
	if (nbits <= 0) {
		*ip = x;
		return 0.0;
	}
	if (nbits <= 32)
		y.b.f2 &= ~((1L << nbits) - 1);
	else {
		y.b.f2 = 0;
		y.b.f1 &= ~((1L << (nbits - 32)) - 1);
	}
	*ip = y.d;
	return x - y.d;
}

